/*
 * mpi.cpp
 *
 *  Created on: Jan 26, 2017
 *      Author: m.toloo
 */


#include "mpi.h"
#include <iostream>
#include <fstream>
using namespace std;

int *read_matrix(string file_name, int *row_num, int *col_num) {
	int* a;
	ifstream file_a(file_name.c_str());
	file_a >> *row_num;
	file_a >> *col_num;
	cout << "Reading " << file_name << " containing " << *row_num << " * " << *col_num << " items..." << endl;
	a = new int[*row_num * *col_num];
	for (int i = 0; i < *row_num; i++) {
		for (int j = 0; j < *col_num; j++) {
			file_a >> a[i * *row_num + j];
		}
	}
        return a;	
}

int main (int argc, char *argv[])
{
	int	process_id, total_process;
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &process_id);
	MPI_Comm_size(MPI_COMM_WORLD, &total_process);

	if (total_process < 2) {
		cout << "At least two processor must define. Run this app with mpirun -n4" << endl;
		MPI_Abort(MPI_COMM_WORLD, process_id);
		return 1;
	}

	int *a, *b, *c, *r, *r_a;
	int a_row_num, a_col_num;
	int b_row_num, b_col_num;
        MPI_Status mpi_status;
        int process_quota;

	if (process_id == 0) {
		r_a = read_matrix("a.dat", &a_row_num, &a_col_num);
		b = read_matrix("b.dat", &b_row_num, &b_col_num);
	    cout << "Sending data to " << total_process << " clients" << endl;
    }
	MPI_Bcast(&a_row_num, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&a_col_num, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&b_row_num, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&b_col_num, 1, MPI_INT, 0, MPI_COMM_WORLD);
	process_quota = a_row_num / (total_process);
	a = new int[process_quota * a_col_num];
    if (process_id != 0)
		b = new int[b_row_num * b_col_num];
	
	MPI_Scatter(r_a, (process_quota * a_col_num), MPI_INT, 
				a, (process_quota * a_col_num), MPI_INT,
				0, MPI_COMM_WORLD);
	MPI_Bcast(b, b_row_num * b_col_num, MPI_INT, 0, MPI_COMM_WORLD);

	printf("%d: Calculating a * b (size=%d)...\n", process_id, process_quota);
	c = new int[process_quota * b_col_num];
	for (int i = 0 ; i < process_quota; i++) {
		for (int j = 0 ; j < b_col_num ; j++) {
			int sum = 0;
			for (int k = 0 ; k < a_col_num ; k++)
			{
//				printf("%d; c[%d] :: a[%d]=%d , b[%d]=%d\n", process_id, i * a_col_num + j, 
	//				i * a_row_num + k, a[i * a_row_num + k], k * b_col_num + j + process_id, b[k * b_col_num + j + process_id]);
				sum += a[i * a_row_num + k] * b[k * b_col_num + j + process_id];
			}
			c[i * a_col_num + j] = sum;
			//cout << process_id << " / " << i * a_col_num + j << "=" << c[ i * a_col_num + j] << endl;
		}
	}
        
	if (process_id == 0)
		r = new int[a_row_num * b_col_num];
		
    MPI_Gather(c, process_quota * a_col_num, MPI_INT, 
               r, process_quota * a_col_num, MPI_INT, 
               0, MPI_COMM_WORLD);
        //cout << process_id << " calculated c: ";
       // for (int i = 0 ; i < process_quota * a_col_num; i++)
        //        cout << process_id << " \\ " << i << " = " << c[i] << " ";
        //cout << endl;
//        MPI_Send(&c, (process_quota * a_col_num), MPI_INT, 0, 10, MPI_COMM_WORLD);
//       if (process_id == 0) {
//            r = new int[a_row_num * b_col_num];
//            for (int pId = 0 ; pId < total_process; pId++) {
//                cout << " Reciving from " << pId << " from byte " << pId * process_quota * a_col_num << endl;
//                MPI_Recv(r + pId * process_quota * a_col_num, process_quota * a_col_num, MPI_INT, pId, 10, MPI_COMM_WORLD, &mpi_status);
//            }
//            cout << " Final result: "<< endl;
	if (process_id == 0)
	{
            ofstream file_r("c.dat");
            for (int i = 0 ; i < a_row_num ; i++)
            {
                for (int j = 0 ; j < b_col_num; j++) {
//                    cout << r[i * a_row_num + j] << ' ' ;
                    file_r << r[i * a_row_num + j] << ' ' ;
                }
//                cout << endl;
                file_r << endl;
            }
   }
//	ofstream file_out("mat-c.dat");
//	cout << "Result: " << endl;
//	for (int i = 0; i < a_row_num; i++) {
//		for (int j = 0; j < b_col_num; j++) {
//			file_out << c[i][j] << "\t";
//		}
//		file_out << '\n';
//	}
        MPI_Finalize();
	return 0;
}


